unit uCtBDE;

(*
  This unit contains the TCtTableEditor class.  This class is used to restructure existing tables.  The basic
  process is as follows :
    1. Register all the fields that you would like your table to have after the restructure
    2. Register the table that will be restructured
    3. Execute the restructure

  Necessary additions to TCtTableEditor :
  If the tables doesn't exist yet, create a new one.
  Register indexes.
  Optionally pack the table (set TblDesc^.pPack to TRUE or FALSE).

  Notes/Ideas :
  There is no way to create a table.  To create a new table this is what would have to happen :
    1. Register the fields
    2. Assign FFieldDefs to a table's FieldDefs
    3. Set the table's DatabaseName, TableName and TableType
    4. Execute TTable.CreateTable

  There is no way to implicitly rename a field.  If a rename is required is could be done
  after registering the table but before executing the restructure.  This is what would have to happen :
    1. Find the field number in the original table
    2. Find the field number in the new table
    3. pFieldDesc^.iFldNum := original field number
    4. pFieldOps^ := crMODIFY
*)

interface

uses
  SysUtils, DB, DBTables, BDE, Grids, Dialogs;

const
  GV_FIELDROSTERFLDNO = 0;
  GV_FIELDROSTERINVARIENTFLDNO = 1;
  GV_FIELDROSTERFLDNAME = 2;
  GV_FIELDROSTERTYPE = 3;
  GV_FIELDROSTERSIZE = 4;
  GV_FIELDROSTERKEY = 5;

type
  TCtTableEditor = class
    // table properties
    FTable: TTable;
    FFieldDefs: TFieldDefs;
    FIsPackTable: Boolean;
    //
    pFieldDesc: pFLDDesc;
    pFieldOps: pCROpType;
    constructor Create;
    destructor Destroy; override;
    procedure RegisterField(FieldName: String; FieldType: TFieldType; Size: Word);
    function RegisterTable(var Table: TTable): Boolean;
    function Execute: Boolean;
  end;

var
  BDEFieldTypeName: Array[ftUnknown..ftTypedBinary] of String;
  BDEFieldTypeSize: Array[ftUnknown..ftTypedBinary] of Boolean;
  PDX2BDEFieldType: Array[fldPDXCHAR..fldPDXUNICHAR] of TFieldType;

  // general library routines
  function ctGetIndex(Str: String; AR: Array of String): Integer;
  //
  ///procedure ctBDEInitialize;
  procedure ctBDESetFieldOps(FldCount: Integer; OldFieldDesc, NewFieldDesc: pFLDDesc; FieldOps: pCROpType);
  procedure ctBDETableToRoster(var Table: TTable; FieldRoster: TStringGrid);
  procedure ctBDERosterToTable(var Table: TTable; FieldRoster: TStringGrid);
  procedure ctBDERosterToNewTable(TblName: String; FieldRoster: TStringGrid);
  //
  procedure TestTableEditor;

implementation

{ ---------------------------------------------------------
  Library Routines
  ---------------------------------------------------------}
function ctGetIndex(Str: String; AR: Array of String): Integer;
var
  i: Integer;
begin
  for i := low(AR) to high(AR) do if (0 = CompareText(Str, AR[i])) then break;
  if (i <= high(AR)) then result := i else result := -1;
end;

{ ---------------------------------------------------------
  TCtTableEditor
  Used to restructure a table.  The basic process is as follows :
  1. Register all the fields you want in the table
  2. Register the table that you want to restructure
  3. Perform the restructure
 ---------------------------------------------------------- }
constructor TCtTableEditor.Create;
begin
  FFieldDefs := TFieldDefs.Create(nil);
  pFieldDesc := nil;
  pFieldOps := nil;
end;

(*
  Clean up.
*)
destructor TCtTableEditor.Destroy;
begin
  FFieldDefs.Free;
  if (nil <> pFieldDesc) then freemem(pFieldDesc);
  if (nil <> pFieldOps) then freemem(pFieldOps);
end;

(*
  Register the fields that are wanted after the restructure.
*)
procedure TCtTableEditor.RegisterField(FieldName: String; FieldType: TFieldType; Size: Word);
begin
  FFieldDefs.Add(FieldName, FieldType, Size, FALSE);
end;

(*
  Table must be opened exclusively.  Changes necessary to restructure table to
  reflect the registered structure are determined.
*)
function TCtTableEditor.RegisterTable(var Table: TTable): Boolean;
var
  FField: TField;
  pField: pFLDDesc;
  pOp: pCROpType;
  Props: CURProps;
  i, j: Integer;
begin
  FTable := Table;
  // Make sure the table is open exclusively so we can get the db handle...
  if NOT(Table.Active) then raise EDatabaseError.Create('Table must be opened to register');
  if NOT(Table.Exclusive) then raise EDatabaseError.Create('Table must be opened exclusively to register');
  // Make sure the table is either Paradox or dBASE...
  Check(DbiSetProp(hDBIObj(Table.Handle), curxltMODE, integer(xltNONE)));
  Check(DbiGetCursorProps(Table.Handle, Props));
  if NOT(Props.szTableType = szPARADOX) then raise EDatabaseError.Create('Only Paradox tables may be registered');
  // okay so far, lets see if any fields need to be added, modified or copied (moved).
  if (nil <> pFieldDesc) then freemem(pFieldDesc);
  if (nil <> pFieldOps) then freemem(pFieldOps);
  pFieldDesc := AllocMem(FFieldDefs.Count * sizeof(FLDDesc));
  pFieldOps := AllocMem(FFieldDefs.Count * sizeof(CROpType));
  try
    FillChar(pFieldDesc^, FFieldDefs.Count * sizeof(FLDDesc), #0);
    FillChar(pFieldOps^, FFieldDefs.Count * sizeof(CROpType), #0);
    pField := pFieldDesc;
    pOp := pFieldOps;
    for i := 0 to FFieldDefs.Count -1 do begin
      with pField^ do begin
        strcopy(szName, PChar(FFieldDefs[i].Name));
        for j := fldPDXCHAR to fldPDXUNICHAR do if (FFieldDefs[i].DataType = PDX2BDEFieldType[j]) then break;
        if (j > fldPDXUNICHAR) then raise EDatabaseError.Create(format('Invalid Datatype in field %s', [FFieldDefs[i].Name]));
        iFldType := j;
        if BDEFieldTypeSize[FFieldDefs[i].DataType] then iUnits1 := FFieldDefs[i].Size else iUnits1 := 1;
        FField := Table.FindField(FFieldDefs[i].Name);
        // if a new field then crADD else if type or size changed then crMODIFY else if moved then crCOPY.
        if (nil = FField) then begin
          iFldNum := 0;
          pOp^ := crADD;
        end else begin
          iFldNum := FField.FieldNo;
          if (FField.DataType <> FFieldDefs[i].DataType) OR (BDEFieldTypeSize[FFieldDefs[i].DataType]) AND (FField.Size <> FFieldDefs[i].Size) then pOp^ := crModify
          else if (iFldNum <> i+1) then pOp^ := crCOPY;
        end;
      end;
      inc(pField);
      inc(pOp);
    end;
  result := TRUE;
  except
    freemem(pFieldDesc); pFieldDesc := nil;
    freemem(pFieldOps); pFieldOps := nil;
    result := FALSE;
  end;
end;

(*
  Setup a table description and restructure the registered table.
*)
function TCtTableEditor.Execute: Boolean;
var
  TableDesc: CRTblDesc;
  hDb: hDbiDb;
begin
  try
    FillChar(TableDesc, sizeof(TableDesc), 0);
    Check(DbiGetObjFromObj(hDBIObj(FTable.Handle), objDATABASE, hDBIObj(hDb)));
    StrPCopy(TableDesc.szTblName, FTable.TableName);
    StrPCopy(TableDesc.szTblType, szPARADOX);
    TableDesc.iFldCount := FFieldDefs.Count;
    Tabledesc.pfldDesc := pFieldDesc;
    TableDesc.pecrFldOp := pFieldOps;
    FTable.Close;
    Check(DbiDoRestructure(hDb, 1, @TableDesc, nil, nil, nil, FALSE));
    result := TRUE;
  except
    result := FALSE;
  end;
end;

(*
  Set the Field Ops to restructure a table.
  FldCount = Number of field in the new table
  OldFieldDesc = Original field description.
  NewFieldDesc = New field description.
  Ops = Field operation to perform.
*)
procedure ctBDESetFieldOps(FldCount: Integer; OldFieldDesc, NewFieldDesc: pFLDDesc; FieldOps: pCROpType);
var
  pOldFld, pNewFld: pFldDesc;
  pOp: pCROpType;
  i: Integer;
begin
  pNewFld := NewFieldDesc;
  pOp := FieldOps;
  for i := 0 to FldCount-1 do begin
    if (0 = pNewFld.iFldNum) then pOp^ := crADD
    else begin
      pOldFld := OldFieldDesc;
      inc(pOldFld, pNewFld.iFldNum);
      if (0 <> StrIComp(pOldFld^.szName, pNewFld^.szName))
      OR (pOldFld.iFldType <> pNewFld.iFldType)
      OR (pOldFld.iUnits1 <> pNewFld.iUnits1) then pOp^ := crMODIFY
      else if (pOldFld.iFldNum <> pNewFld.iFldNum) then pOp^ := crCOPY;
    end;
    inc(pNewFld);
    inc(pOp);
  end;
end;

(*
  Update FieldRoster from a TTable.
  Displays the field position, invarient id, name, type and size of each field in a grid.
*)
procedure ctBDETableToRoster(var Table: TTable; FieldRoster: TStringGrid);
var
  pField, pFields: pFLDDesc;
  Props: CURProps;
  i: Integer;
begin
  // Make sure the table is open exclusively so we can get the db handle...
  if NOT(Table.Active) then raise EDatabaseError.Create('Table must be opened to restructure');
  if NOT(Table.Exclusive) then raise EDatabaseError.Create('Table must be opened exclusively to restructure');
  Check(DbiSetProp(hDBIObj(Table.Handle), curxltMODE, integer(xltNONE)));

  // Get the table properties to determine table type...
  Check(DbiGetCursorProps(Table.Handle, Props));
  // Make sure the table is either Paradox or dBASE...
  if NOT(Props.szTableType = szPARADOX) then raise EDatabaseError.Create('Field altering can only occur on Paradox tables');
  // Allocate memory for the field descriptor...
  pFields := AllocMem(Table.FieldCount * sizeof(FLDDesc));
  try
    // Fill the field descriptor with the existing field information...
    Check(DbiGetFieldDescs(Table.Handle, pFields));
    pField := pFields;
    with FieldRoster do begin
      Row := FixedRows;
      Col := FixedCols;
      RowCount := FixedRows + Table.FieldCount;
      for i := 0 to Table.FieldCount-1 do begin
        with pField^ do begin
          if Table.Fields[i].IsIndexField then Cells[GV_FIELDROSTERKEY, FixedRows+i] := '*';
          Cells[GV_FIELDROSTERFLDNO, FixedRows+i] := IntToStr(i+1);
          Cells[GV_FIELDROSTERINVARIENTFLDNO, FixedRows+i] := IntToStr(iFldNum);
          Cells[GV_FIELDROSTERFLDNAME, FixedRows+i] := szName;
          Cells[GV_FIELDROSTERTYPE, FixedRows+i] := BDEFieldTypeName[PDX2BDEFieldType[iFldType]];
          Cells[GV_FIELDROSTERSIZE, FixedRows+i] := IntToStr(iUnits1);
        end;
        inc(pField);
      end;
    end;
  finally
    if (nil <> pFields) then FreeMem(pFields);
  end;
end;

(*
  Update a Field Roster from a TTable.
*)
procedure ctBDERosterToTable(var Table: TTable; FieldRoster: TStringGrid);
var
  TableDesc: CRTblDesc;
  hDb: hDbiDb;
  pField, pOldFields, pFields: pFLDDesc;
  pFieldOps: pCROpType;
  Props: CURProps;
  i, j: Integer;
begin
  // Make sure the table is open exclusively so we can get the db handle...
  if NOT(Table.Active) then raise EDatabaseError.Create('Table must be opened to restructure');
  if NOT(Table.Exclusive) then raise EDatabaseError.Create('Table must be opened exclusively to restructure');
  Check(DbiSetProp(hDBIObj(Table.Handle), curxltMODE, integer(xltNONE)));

  // Get the table properties to determine table type...
  Check(DbiGetCursorProps(Table.Handle, Props));
  // Make sure the table is either Paradox or dBASE...
  if NOT(Props.szTableType = szPARADOX) then raise EDatabaseError.Create('Field altering can only occur on Paradox tables');
  // Allocate memory for the field descriptor...
  with FieldRoster do begin
    pOldFields := AllocMem(Table.FieldCount * sizeof(FLDDesc));
    pFields := AllocMem((RowCount-FixedRows) * sizeof(FLDDesc));
    pFieldOps := AllocMem((RowCount-FixedRows) * sizeof(CROpType));
    try
      FillChar(pOldFields^, (Table.FieldCount) * sizeof(FLDDesc), #0);
      FillChar(pFields^, (RowCount-FixedRows) * sizeof(FLDDesc), #0);
      FillChar(pFieldOps^, (RowCount-FixedRows) * sizeof(CROpType), #0);
      pField := pFields;
      for i := 0 to RowCount-FixedRows-1 do begin
        with pField^ do begin
          //if Table.Fields[i].IsIndexField then Cells[GV_FIELDROSTERKEY, FixedRows+i] := '*';
          try
            iFldNum := StrToInt(Cells[GV_FIELDROSTERFLDNO, FixedRows+i]);
          except
            iFldNum := 0;
          end;
          StrlCopy(szName, PChar(Cells[GV_FIELDROSTERFLDNAME, FixedRows+i]), sizeof(szName));
          for j := fldPDXCHAR to fldPDXUNICHAR do if (0 = CompareText(BDEFieldTypeName[PDX2BDEFieldType[j]], Cells[GV_FIELDROSTERTYPE, FixedRows+i])) then break;
          iFldType := j;
          iUnits1 := StrToInt(Cells[GV_FIELDROSTERSIZE, FixedRows+i]);
        end;
        inc(pField);
      end;
      Check(DbiGetFieldDescs(Table.Handle, pOldFields));
      ctBDESetFieldOps((RowCount-FixedRows), pOldFields, pFields, pFieldOps);
      // Blank out the structure...
      FillChar(TableDesc, sizeof(TableDesc), 0);
      //  Get the database handle from the table's cursor handle...
      Check(DbiGetObjFromObj(hDBIObj(Table.Handle), objDATABASE, hDBIObj(hDb)));
      // Put the table name in the table descriptor...
      StrPCopy(TableDesc.szTblName, Table.TableName);
      // Put the table type in the table descriptor...
      StrPCopy(TableDesc.szTblType, Props.szTableType);
      // Close the table so the restructure can complete...
      TableDesc.iFldCount := (RowCount-FixedRows);
      Tabledesc.pfldDesc := pFields;
      TableDesc.pecrFldOp := pFieldOps;
      Table.Close;
      ///TableDesc.bPack := FIsPackTable;
      Check(DbiDoRestructure(hDb, 1, @TableDesc, nil, nil, nil, FALSE));
      Table.Open;
    finally
      if (nil <> pOldFields) then FreeMem(pOldFields);
      if (nil <> pFields) then FreeMem(pFields);
      if (nil <> pFieldOps) then FreeMem(pFieldOps);
    end;
  end;
end;

(*
  Create a new table from the field roster
*)
procedure ctBDERosterToNewTable(TblName: String; FieldRoster: TStringGrid);
var
  DB: TDatabase;
  TableDesc: CRTblDesc;
  pField, pFields: pFLDDesc;
  i, j: Integer;
begin
  with FieldRoster do begin
    DB := TDatabase.Create(nil);
    pFields := AllocMem((RowCount-FixedRows) * sizeof(FLDDesc));
    try
      DB.Params.Add(format('PATH = "%s"', [ExtractFilePath(TblName)]));
      DB.Drivername := 'STANDARD';
      DB.DatabaseName := 'CTBDETEMPDB'; // any unused name
      DB.Connected := TRUE;
      FillChar(pFields^, (RowCount-FixedRows) * sizeof(FLDDesc), #0);
      pField := pFields;
      for i := 0 to RowCount-FixedRows-1 do begin
        with pField^ do begin
          //if Table.Fields[i].IsIndexField then Cells[GV_FIELDROSTERKEY, FixedRows+i] := '*';
          StrlCopy(szName, PChar(Cells[GV_FIELDROSTERFLDNAME, FixedRows+i]), sizeof(szName));
          for j := fldPDXCHAR to fldPDXUNICHAR do if (0 = CompareText(BDEFieldTypeName[PDX2BDEFieldType[j]], Cells[GV_FIELDROSTERTYPE, FixedRows+i])) then break;
          iFldType := j;
          iUnits1 := StrToInt(Cells[GV_FIELDROSTERSIZE, FixedRows+i]);
        end;
        inc(pField);
      end;
      FillChar(TableDesc, sizeof(TableDesc), 0);
      StrPCopy(TableDesc.szTblName, TblName);
      StrPCopy(TableDesc.szTblType, szPARADOX);
      TableDesc.iFldCount := (RowCount-FixedRows);
      Tabledesc.pfldDesc := pFields;
      Check(DbiCreateTable(DB.Handle, TRUE, TableDesc));
    finally
      if (nil <> pFields) then FreeMem(pFields);
      if (nil <> DB) then DB.Free;
    end;
  end;
end;

procedure TestTableEditor;
var
  TblEditor: TCtTableEditor;
  Table: TTable;
begin
  Table := TTable.Create(nil);
  TblEditor := TCtTableEditor.Create;
  try
    Table.Exclusive := TRUE;
    Table.TableName := 'D:\D2\BDE\TEST.DB';
    Table.Active := TRUE;
    //
    with TblEditor do begin
      RegisterField('Key', ftAutoInc, 0);
      RegisterField('Name1', ftString, 32);
      RegisterField('Name2', ftString, 32);
      RegisterField('City', ftString, 16);
      RegisterField('State', ftString, 2);
      RegisterField('Zip', ftString, 10);
      RegisterField('Url1', ftString, 40);
      RegisterField('Url2', ftString, 40);
      RegisterField('Phone1', ftString, 16);
      RegisterField('Phone2', ftString, 16);
      RegisterTable(Table);
      Execute;
    end;
  finally
    Table.Free;
    TblEditor.Free;
  end;
end;

Initialization;
  BDEFieldTypeName[ftUnknown] := 'Unknown';
  BDEFieldTypeName[ftString] := 'String';
  BDEFieldTypeName[ftSmallint] := 'Smallint';
  BDEFieldTypeName[ftInteger] := 'Integer';
  BDEFieldTypeName[ftWord] := 'Word';
  BDEFieldTypeName[ftBoolean] := 'Boolean';
  BDEFieldTypeName[ftFloat] := 'Float';
  BDEFieldTypeName[ftCurrency] := 'Currency';
  BDEFieldTypeName[ftBCD] := 'BCD';
  BDEFieldTypeName[ftDate] := 'Date';
  BDEFieldTypeName[ftTime] := 'Time';
  BDEFieldTypeName[ftDateTime] := 'DateTime';
  BDEFieldTypeName[ftBytes] := 'Bytes';
  BDEFieldTypeName[ftVarBytes] := 'VarBytes';
  BDEFieldTypeName[ftAutoInc] := 'AutoInc';
  BDEFieldTypeName[ftBlob] := 'Blob';
  BDEFieldTypeName[ftMemo] := 'Memo';
  BDEFieldTypeName[ftGraphic] := 'Graphic';
  BDEFieldTypeName[ftFmtMemo] := 'FmtMemo';
  BDEFieldTypeName[ftParadoxOle] := 'ParadoxOle';
  BDEFieldTypeName[ftDBaseOle] := 'DBaseOle';
  BDEFieldTypeName[ftTypedBinary] := 'TypedBinary';
  //
  BDEFieldTypeSize[ftUnknown] := TRUE;
  BDEFieldTypeSize[ftString] := TRUE;
  BDEFieldTypeSize[ftSmallint] := FALSE;
  BDEFieldTypeSize[ftInteger] := FALSE;
  BDEFieldTypeSize[ftWord] := FALSE;
  BDEFieldTypeSize[ftBoolean] := FALSE;
  BDEFieldTypeSize[ftFloat] := FALSE;
  BDEFieldTypeSize[ftCurrency] := FALSE;
  BDEFieldTypeSize[ftBCD] := TRUE;
  BDEFieldTypeSize[ftDate] := FALSE;
  BDEFieldTypeSize[ftTime] := FALSE;
  BDEFieldTypeSize[ftDateTime] := FALSE;
  BDEFieldTypeSize[ftBytes] := TRUE;
  BDEFieldTypeSize[ftVarBytes] := TRUE;
  BDEFieldTypeSize[ftAutoInc] := FALSE;
  BDEFieldTypeSize[ftBlob] := TRUE;
  BDEFieldTypeSize[ftMemo] := TRUE;
  BDEFieldTypeSize[ftGraphic] := TRUE;
  BDEFieldTypeSize[ftFmtMemo] := TRUE;
  BDEFieldTypeSize[ftParadoxOle] := TRUE;
  BDEFieldTypeSize[ftDBaseOle] := TRUE;
  BDEFieldTypeSize[ftTypedBinary] := TRUE;
  //
  PDX2BDEFieldType[fldPDXCHAR] := ftString;
  PDX2BDEFieldType[fldPDXNUM] := ftFloat;
  PDX2BDEFieldType[fldPDXMONEY] := ftCurrency;
  PDX2BDEFieldType[fldPDXDATE] := ftDate;
  PDX2BDEFieldType[fldPDXSHORT] := ftSmallint;
  PDX2BDEFieldType[fldPDXMEMO] := ftMemo;
  PDX2BDEFieldType[fldPDXBINARYBLOB] := ftBlob;
  PDX2BDEFieldType[fldPDXFMTMEMO] := ftFmtMemo;
  PDX2BDEFieldType[fldPDXOLEBLOB] := ftParadoxOle;
  PDX2BDEFieldType[fldPDXGRAPHIC] := ftGraphic;
  PDX2BDEFieldType[fldPDXBLOB] := ftBlob;
  PDX2BDEFieldType[fldPDXLONG] := ftInteger;
  PDX2BDEFieldType[fldPDXTIME] := ftTime;
  PDX2BDEFieldType[fldPDXDATETIME] := ftDateTime;
  PDX2BDEFieldType[fldPDXBOOL] := ftBoolean;
  PDX2BDEFieldType[fldPDXAUTOINC] := ftAutoInc;
  PDX2BDEFieldType[fldPDXBYTES] := ftBytes;
  PDX2BDEFieldType[fldPDXBCD] := ftBCD;
  //PDX2BDEFieldType[fldPDXUNICHAR] := 0;
end.

end.
